
(********* ENVIRONMENT "module":

    A  denotable value  is an int loc or a closure.
    An  environment     is a list of  (name, denotable value)  pairs.
    Examples:  
      val env0 = ("y", Loc(1))
                 :: ("x", Loc(0))
                 :: nil                    
      val env1 = ("p", Closure(["a","b"],
                               [Assign("x", Add(Deref("x"), Num(3)))],
                               env0) 
                 :: env0

        Now,  env1  holds bindings for proc  p  and vars  y and x.
        Notice how  p's closure holds (a link to)  env0.

    ML has a  _structure_  construction, and the coding of the environment
    can/should be placed in its own structure file.  Maybe some day....
**********)


datatype Denotable = 
    Loc of int  
 |  Closure of (Id list)* (Comm list) * ((Id * Denotable) list)
     (* NOTE:  See the next line!  *)

type Env = (Id * Denotable) list


(* string-conversion functions:  Denotable2string : Denotable -> string 
                                 Env2string : Env -> string
*)
fun Denotable2string (Loc(v)) = "Loc(" ^ Int.toString(v) ^ ")"
|   Denotable2string (Closure(params, body, link)) =
        "Closure(params:" ^ IList2string(params) ^ ", body:" ^ CommList2string(body)
         ^ ", linkto:"   ^ Env22string(link) ^ ")" 
and Env22string nil = "nil"
|   Env22string ((i,d)::t) =  "("^ i ^ ", " ^ Denotable2string(d) ^ ") :: " ^ (Env22string t)

fun Env2string nil = "nil"
|   Env2string ((i,d)::t) =  "("^ i ^ ", " ^ Denotable2string(d) ^ ")\n:: " ^ (Env2string t)


(* starting environment: *)
val emptyenv : Env = nil

(* error message *)
exception EnvLookupError


(* lookup : Id * Env -> Denotable
   lookup(x, env)  searches for  x  in  env  and returns the denotable value
   paired with it,  e.g.,
     lookup("b",  ("a", Loc(0)) :: ("b", Loc(1) :: nil))  returns  Loc(1).
   Raises  EnvLookupError  when  x  is not found in  env. 
*)
fun lookup(x, nil: Env) =  raise EnvLookupError
|   lookup(x, ((y,d)::s)) =  if x = y
                             then d
                             else lookup(x,s)


(* declare : Id * Denotable * Env -> Env
   declare(x, d, env)  returns an Env that has (x,d) attached to env.
*)
fun declare(x, d, env: Env) = (x, d):: env

