
(*************** STORE "module": 

   defines linear storage as a list of cells that hold ints.

   Example:   [5, 7, 9]   models a store that holds  5 in location 0,
      7 in location 1, and 9 in location 2.
      
****************)

type Store = int list

(* Store2string : Store  string
   string-conversion function for printing a store
*)
(*fun int2string n = 
      let fun digit2string n = implode[chr(n + ord(#"0"))]
          fun i2s n =
                if n > 0 
                then (i2s (n div 10))^(digit2string(n mod 10))
                else ""
       in if n = 0 then "0"
          else if n > 0 then i2s n
          else (* negative *) "~"^(i2s (~n))
       end
*)

fun Store2string (mem: Store) =
    let fun s2s [] = ""
        |   s2s (v::s) =  Int.toString(v) ^ ", " ^ (s2s s)
    in
    "[ " ^ (s2s mem) ^ "]"
    end


(* initial, empty store value: *)
val emptymem: Store = []


(* error messages: *)
exception StoreFetchError
exception StoreError
exception RevertError


(* allocCell : Store -> int * Store
   allocCell mem   builds a store that is  mem   extended by a new cell,
   initialized to 0, e.g.,   allocCell [4,5,7]  returns the pair,
     (3, [4,5,7,0]),  namely, the new location number and the updated store.
   NOTE:  in ML,  length mem  returns the integer length of list  mem.
*)
fun allocCell mem = 
       let val newloc = length mem
       in  (newloc, mem @ [0])
       end
    
   
(* fetch : int * Store -> int
   fetch(i, mem)  returns the int saved in location  i
     in  mem,  e.g.,  fetch(2, [4,5,7,9])  returns  7
   Raises  StoreFetchError  if location  i  not found in  mem 
*)
fun fetch(i, []: Store) = raise StoreFetchError
|   fetch(0, v::s) = v
|   fetch (n, v::s) = fetch ((n-1),s)
;

(* update : int * int * Store -> Store
   update(i, v, mem)  returns a store that looks like  mem  with the int
   at location  i  replaced by  v.
   E.g.,  update(1, 4, [3,5,7,9])  returns  [3,4,7,9]
   Raises  StoreError  if  location  i  not found in  mem 
*)
fun update(i, v, s: Store) = 
    (print "update not yet implemented\n";
     s)
;

(* revert : int * Store -> Store
   revert(i, mem)  returns a store that is  mem  shortened to length  i,
   e.g.,  revert(2, [3,5,7,9])  returns  [3,5].
   Raises  RevertError  if  i > length(mem).
   NOTE: in ML,  length mem   returns the int length of list  mem
         (see how it's used in the  allocCell  function above) 
*)
fun revert(n, s: Store) =
    (print "revert not yet implemented\n";
     s)
;

